#include "mex.h"
#include "fft_evd.c"
#include <stdio.h>
#include <math.h>

int rnd(double number);

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
    
	long int sampI, probeI, gI, probeStart, probeEnd,probeSize, logSize;
	int	   i;
    int    numProbes, numSamps, numKernelSamps;
	double kernelT, kernelOffset;
	double *probeLoc, *profile, *mapLoc, *probeRange, *g, *kernel1, *R, *kernel2, *ProfileS;
	double (*kernelWeights)[2];
	double (*RWeights)[2];
	double (*Rfft)[2];
	double (*Wfft)[2];
	double (*Pfft)[2];
	
	double *profileOut;
    double (*profileC)[2];
    double (*ProfileC)[2];
   
/*	Read input parameters */
    numProbes		= mxGetN(prhs[0]);
    probeLoc		= mxGetPr(prhs[0]);
	profile			= mxGetPr(prhs[1]);
	numSamps		= mxGetN(prhs[2]);
	mapLoc			= mxGetPr(prhs[2]);
	probeRange		= mxGetPr(prhs[3]);
	numKernelSamps	= mxGetN(prhs[4]);
	g				= mxGetPr(prhs[4]);
	kernel1			= mxGetPr(prhs[5]);
	R				= mxGetPr(prhs[6]);
	kernel2			= mxGetPr(prhs[7]);
	
	kernelT			= (double)(mxGetScalar(prhs[8]));
	kernelOffset	= (double)(mxGetScalar(prhs[9]));

/*	Create outout parameters */
	plhs[0]			= mxCreateDoubleMatrix(1, numSamps, mxREAL);
	ProfileS		= mxGetPr(plhs[0]);



	
	
/*	Convolution body
	//printf("numProbes     : %d\n", numProbes);
//	printf("numSamps      : %d\n", numSamps);
//	printf("numKernelSamps: %d\n\n", numKernelSamps);
//	printf("kernelT		  : %f\n", kernelT);
//	printf("kernelOffset  : %f\n\n", kernelOffset);
//	printf("probeLoc      : %f, %f, ..., %f\n", probeLoc[0], probeLoc[1], probeLoc[numProbes-1]);
//	printf("profile       : %f, %f, ..., %f\n", profile[0], profile[1], profile[numProbes-1]);
//	printf("mapLoc        : %f, %f, ..., %f\n\n", mapLoc[0], mapLoc[1], mapLoc[numSamps-1]);
//	printf("probeRange    : %f, %f, ..., %f\n", probeRange[0], probeRange[2], probeRange[2*numSamps-1-1]);
//	printf("                %f, %f, ..., %f\n\n", probeRange[1], probeRange[3],  probeRange[2*numSamps-1]);
//	printf("g             : %f, %f, ..., %f\n", g[0], g[1], g[numKernelSamps-1]);
//	printf("kernel        : %e, %e, ..., %e\n", kernel[0], kernel[1], kernel[numKernelSamps-1]); */
	
	for (sampI = 0; sampI < numSamps; sampI++) {
		ProfileS[sampI] = 0.0;
		
		probeStart = (long int)(rnd(probeRange[2*sampI]-1));
		probeEnd   = (long int)(rnd(probeRange[2*sampI+1]));
		probeSize  = probeEnd - probeStart;
		logSize    = ceil(log2((double)probeSize));
		probeSize  = (long int)(pow(2,logSize));
		kernelWeights	= calloc(4*probeSize, sizeof(double));
		RWeights		= calloc(4*probeSize, sizeof(double));
		Wfft			= calloc(4*probeSize, sizeof(double));
		Rfft			= calloc(4*probeSize, sizeof(double));
		Pfft			= calloc(4*probeSize, sizeof(double));
		
		for (probeI = probeStart, i = 0; probeI < probeEnd; probeI++, i++) {
			gI = (long int)(rnd((probeLoc[probeI] -  mapLoc[sampI] + kernelOffset) / kernelT));
			kernelWeights[i][0]				= kernel1[gI] * profile[probeI];
			RWeights[i][0]					= R[i];
			RWeights[2*probeSize-i-1][0]	= R[i+1];
		}
		
		/*if (sampI == 2) {
//			printf("\n");
//			printf("\n");
//			printf("KWeights[0] = %f, %f, %f, %f, ..., %f, ..., %f, %f, %f\n", kernelWeights[0][0],kernelWeights[1][0],kernelWeights[2][0],kernelWeights[3][0], kernelWeights[probeSize][0],kernelWeights[2*probeSize-3][0],kernelWeights[2*probeSize-2][0],kernelWeights[2*probeSize-1][0]);
//			printf("KWeights[1] = %f, %f, %f, %f, ..., %f, ..., %f, %f, %f\n", kernelWeights[0][1],kernelWeights[1][1],kernelWeights[2][1],kernelWeights[3][1], kernelWeights[probeSize][1],kernelWeights[2*probeSize-3][1],kernelWeights[2*probeSize-2][1],kernelWeights[2*probeSize-1][1]);
//			printf("\n");
//			printf("RWeights[0] = %f, %f, %f, %f, ..., %f, ..., %f, %f, %f\n", RWeights[0][0],RWeights[1][0],RWeights[2][0],RWeights[3][0], RWeights[probeSize][0],RWeights[2*probeSize-3][0],RWeights[2*probeSize-2][0],RWeights[2*probeSize-1][0]);
//			printf("RWeights[1] = %f, %f, %f, %f, ..., %f, ..., %f, %f, %f\n", RWeights[0][1],RWeights[1][1],RWeights[2][1],RWeights[3][1], RWeights[probeSize][1],RWeights[2*probeSize-3][1],RWeights[2*probeSize-2][1],RWeights[2*probeSize-1][1]);
//		}*/
		
		fft(2*probeSize, kernelWeights, Wfft);
		fft(2*probeSize, RWeights, Rfft);
		
		for (i = 0; i < 2*probeSize; i++) {
			Pfft[i][0] = Wfft[i][0]*Rfft[i][0] + Wfft[i][1]*Rfft[i][1];
			Pfft[i][1] = Wfft[i][1]*Rfft[i][0] - Wfft[i][0]*Rfft[i][1];
		}
		
		ifft(2*probeSize, kernelWeights, Pfft);
		
		/*if (sampI == 2) {
//			printf("\n");
//			printf("KConv[0] = %f, %f, %f, %f, ..., %f, ..., %f, %f, %f\n", kernelWeights[0][0],kernelWeights[1][0],kernelWeights[2][0],kernelWeights[3][0], kernelWeights[probeSize][0],kernelWeights[2*probeSize-3][0],kernelWeights[2*probeSize-2][0],kernelWeights[2*probeSize-1][0]);
//			printf("KConv[1] = %f, %f, %f, %f, ..., %f, ..., %f, %f, %f\n", kernelWeights[0][1],kernelWeights[1][1],kernelWeights[2][1],kernelWeights[3][1], kernelWeights[probeSize][1],kernelWeights[2*probeSize-3][1],kernelWeights[2*probeSize-2][1],kernelWeights[2*probeSize-1][1]);
//		}*/
		
		ProfileS[sampI] = 0.0;
		for (probeI = probeStart, i = 0; probeI < probeEnd; probeI++, i++) {
			gI = (long int)(rnd((probeLoc[probeI] -  mapLoc[sampI] + kernelOffset) / kernelT));
			ProfileS[sampI] += kernel2[gI] * profile[probeI] * kernelWeights[i][0];
		}
		
		free(kernelWeights);
		free(RWeights);
		free(Wfft);
		free(Rfft);
		free(Pfft);
	}
	
    	
/*     Get Vector
//     nbi    = mxGetN(prhs[0]);
//     profil = mxGetPr(prhs[0]); 

//     Get scaler
//     Kmaxi = (int)(mxGetScalar(prhs[1]));
    
//     Get string
//     filePathLength = mxGetN(prhs[2]) + 1;
//     filePath = (char *)mxCalloc(filePathLength, sizeof(char));
//     mxGetString(prhs[2], filePath, filePathLength); 
//     mxFree(filePath);
    
    
//     save double matrix        
//     plhs[0] = mxCreateDoubleMatrix(Kmaxi, Kmaxi, mxREAL);
//     breakMatrix = mxGetPr(plhs[0]);*/
    
    
    return;
}

int rnd(double number)
{
  return (number >= 0) ? (int)(number + 0.5) : (int)(number - 0.5);
}
