function [data, fieldParams, convInfo, conv] = getScaleSpace(data, kws, kernelFile, fieldParams)
% 
    printf(-1,1,'Performing kernel convolutions:\n')
    numKernels = length(kws);

    [convInfo] = doPlatformPreproc(data.mapLoc, kernelFile, kws);
    
    
    if (fieldParams.segFlag == 1) && (fieldParams.sigSys ~= 0)
            [data, fieldParams, convInfo] = calcSegSysCovariance(data, fieldParams, convInfo);
    end

    printf(-1,1,'\tConvolving... Kernel index (of %d) = ', numKernels);
    printf(-1,1,'');
    

    conv = [];
    conv.kw = cell(1,numKernels);
    for kwI = 1:numKernels
        conv.kw{kwI}.lrSmooth       = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.muSmooth       = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.zSmooth        = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.K_R_K          = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.sys_K_R_K     = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.dK_R_dK        = zeros(size(convInfo.kw{kwI}.mapLoc));     
        conv.kw{kwI}.sys_dK_R_dK     = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.K_R_dK         = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.sys_K_R_dK     = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.V              = zeros(size(convInfo.kw{kwI}.mapLoc));
        conv.kw{kwI}.sampI      = logical(ones(size(convInfo.kw{kwI}.mapLoc)));
        conv.kw{kwI}.zeroI      = logical(zeros(size(convInfo.kw{kwI}.mapLoc)));
    end

    probeI = logical(ones(1, data.P));

    for kwI = 1:numKernels
        printf(1, 1, '%d\t', kwI);
        conv.kw{kwI}.lrSmooth   = convolve(conv.kw{kwI}.lrSmooth, conv.kw{kwI}.sampI, data.lrAgr, probeI, convInfo, kws(kwI), 0);
        conv.kw{kwI}.muSmooth   = convolve(conv.kw{kwI}.muSmooth, conv.kw{kwI}.sampI, ones(1, data.P), probeI, convInfo, kws(kwI), 0);
        
        conv.kw{kwI}.K_R_K      = quadProdAutoCorr(conv.kw{kwI}.K_R_K, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), fieldParams.R, 0, 0);
        conv.kw{kwI}.dK_R_dK = quadProdAutoCorr(conv.kw{kwI}.dK_R_dK, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), fieldParams.R, 1, 1);
        conv.kw{kwI}.K_R_dK = quadProdAutoCorr(conv.kw{kwI}.K_R_dK, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), fieldParams.R, 0, 1);
        
        if (fieldParams.sigSys ~= 0) && (fieldParams.segFlag == 1)
            conv.kw{kwI}.sys_K_R_K          = quadProdSysCov(conv.kw{kwI}.sys_K_R_K, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), fieldParams, 0, 0);
            conv.kw{kwI}.sys_dK_R_dK      = quadProdSysCov(conv.kw{kwI}.sys_K_R_K, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), fieldParams, 1, 1);
            conv.kw{kwI}.sys_K_R_dK        = quadProdSysCov(conv.kw{kwI}.sys_K_R_K, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), fieldParams, 0, 1);
        elseif (fieldParams.sigSys ~= 0) 
            sysR = zeros(size(fieldParams.R));
            sysR(1) = 1;
            conv.kw{kwI}.sys_K_R_K      = quadProdAutoCorr(conv.kw{kwI}.sys_K_R_K, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), sysR, 0, 0);
            conv.kw{kwI}.sys_dK_R_dK = quadProdAutoCorr(conv.kw{kwI}.sys_dK_R_dK, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), sysR, 1, 1);
            conv.kw{kwI}.sys_K_R_dK = quadProdAutoCorr(conv.kw{kwI}.sys_K_R_dK, conv.kw{kwI}.sampI, probeI, convInfo, kws(kwI), sysR, 0, 1);
        end
        
        dK_C_dK = fieldParams.sig^2*conv.kw{kwI}.dK_R_dK + fieldParams.sigSys^2*conv.kw{kwI}.sys_dK_R_dK;
        K_C_dK   = fieldParams.sig^2*conv.kw{kwI}.K_R_dK + fieldParams.sigSys^2*conv.kw{kwI}.sys_K_R_dK;
        K_C_K     = fieldParams.sig^2*conv.kw{kwI}.K_R_K + fieldParams.sigSys^2*conv.kw{kwI}.sys_K_R_K;
        
        conv.kw{kwI}.zSmooth = (conv.kw{kwI}.lrSmooth - fieldParams.mu*conv.kw{kwI}.muSmooth)./(sqrt(K_C_K));
%         conv.kw{kwI}.zSmooth = (conv.kw{kwI}.lrSmooth - fieldParams.mu*conv.kw{kwI}.muSmooth)./(fieldParams.sig*sqrt(conv.kw{kwI}.K_R_K));
%         conv.kw{kwI}.V = conv.kw{kwI}.dK_R_dK ./ conv.kw{kwI}.K_R_K - (conv.kw{kwI}.K_R_dK ./ conv.kw{kwI}.K_R_K).^2;
        conv.kw{kwI}.V = dK_C_dK./ K_C_K - (K_C_dK ./ K_C_K).^2;
        conv.kw{kwI}.V = max(conv.kw{kwI}.V, 0);
        conv.kw{kwI}.numEuler = 1;
        conv.kw{kwI}.intSqrtV    = sum(sqrt(conv.kw{kwI}.V)) * (convInfo.kw{kwI}.mapLocSqsh(2) - convInfo.kw{kwI}.mapLocSqsh(1));
        conv.kw{kwI}.eEuler = getAnalNegLogEEuler(conv.kw{kwI}.intSqrtV, conv.kw{kwI}.numEuler, abs(conv.kw{kwI}.zSmooth)) - log10(2);
    end
    printf(-1,1, '\n');
    


end

function [preConvInfo] = doPlatformPreproc(mapLoc, kernelFile, kws)
    numKernels = length(kws);
    printf(-1, 1, '\tPreprocessing... Kernel index (of %d) = ', numKernels);
    printf(-1,1,'');

    sqshProbPos = cell(1,numKernels);
    sqshSampPos = cell(1,numKernels);
    realSampPos = cell(1,numKernels);

    preConvProbeRanges  = cell(1,numKernels);

    preConvInfo = [];
    preConvInfo.probeLoc = mapLoc;
    preConvInfo.kws = kws;
    preConvInfo.kw = cell(1,numKernels);


    load(kernelFile);
    % Reflect and trim kernels
    kernel      = kernel(end:-1:1);
    d_dg_kernel = d_dg_kernel(end:-1:1);

    cumKernel = cumsum(abs(d_dg_kernel));
    kernelPerc = 0.0001;
    kernelIMin = find(cumKernel < kernelPerc*cumKernel(end));
    kernelIMin = kernelIMin(end);
    kernelIMax = find(cumKernel > (1-kernelPerc)*cumKernel(end));
    kernelIMax = kernelIMax(1);
    g = g(kernelIMin:kernelIMax);
    kernel = kernel(kernelIMin:kernelIMax);
    d_dg_kernel = d_dg_kernel(kernelIMin:kernelIMax);

    preConvInfo.g = g;
    preConvInfo.kernel = kernel;
    preConvInfo.d_dg_kernel = d_dg_kernel;
    for kwI = 1:numKernels
        printf(1, 1, ' %d\t', kwI);
        printf(0, 1, '');
        [sqshProbPos{kwI}, sqshSampPos{kwI}, realSampPos{kwI}] = getSquashPos(mapLoc, kws(kwI));
        [preConvProbeRanges{kwI}, kernelT, kernelTOffset] = preprocConvSTEP1(sqshProbPos{kwI}, sqshSampPos{kwI}, g, kws(kwI));
        printf(2,1,'');
        preConvInfo.kw{kwI}.kernelT = kernelT;
        preConvInfo.kw{kwI}.kernelOffset = kernelTOffset;

        preConvInfo.kw{kwI}.mapLoc       = realSampPos{kwI};
        preConvInfo.kw{kwI}.mapLocSqsh   = sqshSampPos{kwI};
        preConvInfo.kw{kwI}.probeLocSqsh = sqshProbPos{kwI};
        preConvInfo.kw{kwI}.probes       = preConvProbeRanges{kwI};
        
    end
    printf(1,1, '100%%\n');

    probesPos = preConvInfo.probeLoc;
    minPos = min(probesPos);
    maxPos = max(probesPos);
    probesI   = 1:length(preConvInfo.probeLoc);
    for i = 1:numKernels
        smoothPos = preConvInfo.kw{i}.mapLoc;
        smoothPos(smoothPos < minPos) = minPos;
        smoothPos(smoothPos > maxPos) = maxPos;
        nearestI = round(interp1(probesPos, probesI, smoothPos));
        preConvInfo.kw{i}.nearestProbeI = nearestI;    
    end


    for i = 1:numKernels
        smoothPos = preConvInfo.kw{i}.mapLoc;
        minPos = min(smoothPos);
        maxPos = max(smoothPos);
        smoothI = 1:length(preConvInfo.kw{i}.mapLoc);

        probesPos = preConvInfo.probeLoc;
        probesPos(probesPos < minPos) = minPos;
        probesPos(probesPos > maxPos) = maxPos;
        nearestI = round(interp1(smoothPos, smoothI, probesPos));
        preConvInfo.kw{i}.nearestSmoothI = nearestI;    
    end
end


function [probePosRel, smoothSampRel, smoothSampAbs] = getSquashPos(pos, kw)

    sampleT = kw/4; 

    posOld = double(pos);
    diffPos = diff(posOld);
    diffPos(diffPos>6*kw) = 6*kw;
    pos = cat(2, posOld(1), posOld(1) + cumsum(diffPos));
    diffPos = posOld - pos;

    g_samp = pos(1):sampleT:(pos(end)+sampleT);

    posMid = pos(1:end-1) + diff(pos)/2;
    posEdge = cat(2, 0, posMid, g_samp(end)+1);
    [~,fitI] = histc(g_samp, posEdge);
    g_real = g_samp + diffPos(fitI);

    probePosRel   = pos;
    smoothSampRel = g_samp;
    smoothSampAbs = g_real;

end

function [probe_samp, kernelT, kernelOffset] = preprocConvSTEP1(pos, sampPos, g, kw)

    pos = double(pos);
    % plot(kernel)

    sampleT = sampPos(2) - sampPos(1);
    numProbes = length(pos);

    g_middleI = find(g == 0);

    % Scale kernel to kw
    g = kw*g;
    kernelT         = g(2) - g(1);
    kernelOffset    = (g_middleI-1)*kernelT;

    kernelDomainLen = g(end) - g(1) + 1;

    % Calculate the maximum number of probes that can be squashed into sampleT
    % and the Kernel
    pos_shift = zeros(1,length(pos));
    pos_shift(1) = pos(1);
    pos_shift(2:end) = pos(1:end-1);
    pos_diff = pos - pos_shift;
    pos_diff = pos_diff(2:end);
    pos_diff_sort = sort(pos_diff, 'ascend');
    cum_diff_sort = cumsum(pos_diff_sort);
    maxProbesInT        = sum(cum_diff_sort <= sampleT) + 1;
    maxProbesInKernel   = sum(cum_diff_sort <= kernelDomainLen) + 1;

    g_samp = sampPos;

    probe_samp = zeros(2,length(g_samp));

    % currStartProbe = 1;
    currStartProbe = find(pos > (g_samp(1) - kernelDomainLen/2),1);

    for gI = 1:length(g_samp)
        if mod(gI, 1000) == 0
            printf(1,1, '%0.2f%%', 100* gI / length(g_samp));
        end
        relProbes = currStartProbe:min(currStartProbe+maxProbesInT+maxProbesInKernel, numProbes);
    %     relProbes = 1:length(pos)
        relProbeDistances = pos(relProbes) - g_samp(gI);
    %     relProbeDistances = g_samp(gI) - pos(relProbes);
        relProbeMin = find(relProbeDistances >= g(1));
        if ~isempty(relProbeMin)
            relProbeMin = relProbes(relProbeMin(1));
        end
        relProbeMax = find(relProbeDistances <= g(end));
        if ~isempty(relProbeMax)
            relProbeMax = relProbes(relProbeMax(end));
        end


        if (~isempty(relProbeMin)) && (~isempty(relProbeMax))
            probe_samp(:,gI) = [relProbeMin;relProbeMax];
        else
            disp('Error')
            probe_samp(:,gI) = [2,1];
        end

        if ~isempty(relProbeMin)
            currStartProbe = relProbeMin;
        end

    end

end

function [ profileS ] = convolve(profileS, sampI, profile, probeI, preConvInfo, kw, DOrder)

    profile = profile .* probeI;

    g = preConvInfo.g;
    if DOrder == 0
        kernel = preConvInfo.kernel;
    else
        kernel = preConvInfo.d_dg_kernel/kw;
    end

    kwI = find(kw == preConvInfo.kws,1);
    if isempty(kwI)
        disp('Kernel width not defined in preprocessing')
        return;
    end
    kernelT         = preConvInfo.kw{kwI}.kernelT;
    kernelOffset    = preConvInfo.kw{kwI}.kernelOffset;

    probeLoc    = preConvInfo.kw{kwI}.probeLocSqsh;
    mapLoc      = preConvInfo.kw{kwI}.mapLocSqsh;
    probeRange  = preConvInfo.kw{kwI}.probes;


    profileS(sampI) = convolve_c(probeLoc, profile, mapLoc(sampI), probeRange(:,sampI), g, kernel, kernelT, kernelOffset);


end

function [ profileS ] = quadProdAutoCorr(profileS, sampI, probeI, preConvInfo, kw, R, DOrder1, DOrder2)
    profile = double(probeI);
    if length(profile) > length(R)
        disp('R not correct length')
        numExtra = length(profile) - length(R);
        R = cat(2,R(1:round(end/2)), zeros(1,numExtra), R(round(end/2)+1:end));
    end

    g = preConvInfo.g;
    if DOrder1 == 0
        kernel1 = preConvInfo.kernel;
    else
        kernel1 = preConvInfo.d_dg_kernel/kw;
    end

    if DOrder2 == 0
        kernel2 = preConvInfo.kernel;
    else
        kernel2 = preConvInfo.d_dg_kernel/kw;
    end

    kwI = find(kw == preConvInfo.kws,1);

    if isempty(kwI)
        disp('Kernel width not defined in preprocessing')
        return;
    end
    kernelT         = preConvInfo.kw{kwI}.kernelT;
    kernelOffset    = preConvInfo.kw{kwI}.kernelOffset;

    probeLoc    = preConvInfo.kw{kwI}.probeLocSqsh;
    mapLoc      = preConvInfo.kw{kwI}.mapLocSqsh;
    probeRange  = preConvInfo.kw{kwI}.probes;

    profileS(sampI) = quadProdAutoCorr_c(probeLoc, profile, mapLoc(sampI), probeRange(:,sampI), g, kernel1, R, kernel2, kernelT, kernelOffset);
end

function [ profileS ] = quadProdSysCov(profileS, sampI, probeI, preConvInfo, kw, fieldParams, DOrder1, DOrder2)
    profile = double(probeI);
    g = preConvInfo.g;
    if DOrder1 == 0
        kernel1 = preConvInfo.kernel;
    else
        kernel1 = preConvInfo.d_dg_kernel/kw;
    end

    if DOrder2 == 0
        kernel2 = preConvInfo.kernel;
    else
        kernel2 = preConvInfo.d_dg_kernel/kw;
    end

    kwI = find(kw == preConvInfo.kws,1);

    if isempty(kwI)
        disp('Kernel width not defined in preprocessing')
        return;
    end
    kernelT         = preConvInfo.kw{kwI}.kernelT;
    kernelOffset    = preConvInfo.kw{kwI}.kernelOffset;

    probeLoc    = preConvInfo.kw{kwI}.probeLocSqsh;
    mapLoc      = preConvInfo.kw{kwI}.mapLocSqsh;
    probeRange  = preConvInfo.kw{kwI}.probes;

    profileS(sampI) = quadProdSysCov_c(probeLoc, profile, mapLoc(sampI), probeRange(:,sampI), g, kernel1, fieldParams.aggSegI, fieldParams.sysCovMat, kernel2, kernelT, kernelOffset);
end




function [data, fieldParams, convInfo] = calcSegSysCovariance(data, fieldParams, convInfo)
    
    [data, segInfo] = getParams(data, [], [], [], [], 'seg');
    [~, maxSegs] = size(segInfo.segStarts);

    fieldParams.aggSegI = cumsum([0, diff(segInfo.cumIAgg) ~= 0]) + 1;
%     fieldParams.aggSegI = cumsum([0, diff(data.lrAgr) ~= 0]) + 1;
%     figure;
%     hold;
%     plot(fieldParams.aggSegI);
%     plot(fieldParams.aggSegI2, 'r');
    maxCover = 0;
    for kwI = 1:length(convInfo.kws)
        convInfo.kw{kwI}.segProbes = fieldParams.aggSegI(convInfo.kw{kwI}.probes);
        maxCover = max(maxCover, max(convInfo.kw{kwI}.segProbes(2,:) - convInfo.kw{kwI}.segProbes(1,:)+1));
    end
%     maxCover
    
    [~, repProbes] = unique(fieldParams.aggSegI);
    numRepProbes = length(repProbes);
    
    lrSegI = zeros(data.S, numRepProbes);
    
    for sampI = 1:data.S
        
        currNumSegs = find(segInfo.segStarts(sampI, :) == 0, 1, 'first') - 1;
        if isempty(currNumSegs)
            currNumSegs = maxSegs;
        end
        x = [segInfo.segStarts(sampI, 1:currNumSegs), segInfo.segEnds(sampI, currNumSegs)+1];
        y = 1:(currNumSegs+1);
        lrSegI(sampI,:) = floor(interp1(x, y, repProbes));
    end
    
    fieldParams.maxCover = maxCover;
    fieldParams.lrSegI = lrSegI;
    fieldParams.segStarts = segInfo.segStarts;
    fieldParams.segEnds = segInfo.segEnds;
    
    covMatrix = calcSegSysCovariance_c(1, maxCover, lrSegI, segInfo.segStarts, segInfo.segEnds);

    fieldParams.sysCovMat = covMatrix' / (data.S^2);

    
end










