function [ results ] = outputEvents(data,  results, outputFile)

if isempty(results.layerSigns)
    noOutput(outputFile);
    return;
end

numLayers = length(results.layerSigns);

while isempty(results.layerSigns{end})
    numLayers = numLayers - 1;
    results.layerSigns = results.layerSigns(1:end-1);
    results.layerSegs = results.layerSegs(1:end-1);
    results.layerMus = results.layerMus(1:end-1);
    results.layerStd = results.layerStd(1:end-1);
end

if isempty(results.layerSigns)
    noOutput(outputFile);
    return;
end

fp = fopen(outputFile, 'w');

fprintf(fp, 'Event ID\tGain/Loss\tPhysical position\tRecursive level\tNote\n');


for layerI = 1:numLayers
    [~, sortI] = sort(results.layerSegs{layerI}(1,:), 'ascend');
    results.layerSegs{layerI} = results.layerSegs{layerI}(:,sortI);
    results.layerSigns{layerI} = results.layerSigns{layerI}(sortI);
    results.layerMus{layerI} = results.layerMus{layerI}(sortI);
    results.layerStd{layerI} = results.layerStd{layerI}(sortI);
    
    fprintf(fp, '\n'); 
    
    posI = find(results.layerSigns{layerI} == +1);
    numPosSegs = length(posI);
    for eventI = 1:numPosSegs
        if layerI > 1
            posOverlap = find((results.layerSegs{layerI-1}(1,prevLayerPosI) <= results.layerSegs{layerI}(1,posI(eventI))) & (results.layerSegs{layerI-1}(2,prevLayerPosI) >= results.layerSegs{layerI}(2,posI(eventI))));
            negOverlap = find((results.layerSegs{layerI-1}(1,prevLayerNegI) <= results.layerSegs{layerI}(1,posI(eventI))) & (results.layerSegs{layerI-1}(2,prevLayerNegI) >= results.layerSegs{layerI}(2,posI(eventI))));
            if ~isempty(posOverlap)
                note = sprintf('Relative to Gain%02d_%d', posOverlap, layerI-1);
            else
                note = sprintf('Relative to Loss%02d_%d', negOverlap, layerI-1);
            end
        else
            note = '-';
        end
        eventID = sprintf('Gain%02d_%d', eventI, layerI);
        gainLoss = 'Gain';
        probeStart = results.layerSegs{layerI}(1,posI(eventI));
        probeEnd = results.layerSegs{layerI}(2,posI(eventI));
        chrom = data.chrom(probeStart);
        posStart = data.relMapLoc(probeStart);
        posEnd = data.relMapLoc(probeEnd);
        
        physicalPos = sprintf('chr%d:%d-%d (probes %d-%d)', chrom, posStart, posEnd, probeStart, probeEnd);
        
        fprintf(fp, '%s\t%s\t%s\t%d\t%s\n', eventID, gainLoss, physicalPos, layerI, note);
    end
    
    negI = find(results.layerSigns{layerI} == -1);
    numNegSegs = length(negI);
    for eventI = 1:numNegSegs
        if layerI > 1
            posOverlap = find((results.layerSegs{layerI-1}(1,prevLayerPosI) <= results.layerSegs{layerI}(1,negI(eventI))) & (results.layerSegs{layerI-1}(2,prevLayerPosI) >= results.layerSegs{layerI}(2,negI(eventI))));
            negOverlap = find((results.layerSegs{layerI-1}(1,prevLayerNegI) <= results.layerSegs{layerI}(1,negI(eventI))) & (results.layerSegs{layerI-1}(2,prevLayerNegI) >= results.layerSegs{layerI}(2,negI(eventI))));
            if ~isempty(posOverlap)
                note = sprintf('Relative to Gain%02d_%d', posOverlap, layerI-1);
            else
                note = sprintf('Relative to Loss%02d_%d', negOverlap, layerI-1);
            end
        else
            note = '-';
        end
        eventID = sprintf('Loss%02d_%d', eventI, layerI);
        gainLoss = 'Loss';
        probeStart = results.layerSegs{layerI}(1,negI(eventI));
        probeEnd = results.layerSegs{layerI}(2,negI(eventI));
        chrom = data.chrom(probeStart);
        posStart = data.relMapLoc(probeStart);
        posEnd = data.relMapLoc(probeEnd);
        
        physicalPos = sprintf('chr%d:%d-%d (probes %d-%d)', chrom, posStart, posEnd, probeStart, probeEnd);
        
        fprintf(fp, '%s\t%s\t%s\t%d\t%s\n', eventID, gainLoss, physicalPos, layerI, note);
    end
    
    prevLayerPosI = posI;
    prevLayerNegI = negI;
    
end

fclose(fp);


end

function noOutput(outputFile)
    fp = fopen(outputFile, 'w')
    fprintf(fp, 'No significant regions found\n');
    fclose(fp);
end